/*
 * mallory.c
 * Copyright (C) 2005 Robin KAY. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 */

#include "compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <sys/acl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <alloca.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <dlfcn.h>
#include <limits.h>
#include <link.h>
#include <pthread.h>
#include <thread.h>
#include <unistd.h>

#include "utils.h"
#include "mallory.h"
#include "tracker.h"

#define INIT() if (!init_called) do {init_called = 1; init();} while(0)

typedef struct {
  const char *name;
  const void **fptr;
  const enum {REQUIRED = 1, OPTIONAL = 0} required;
} fptrname_t;

static int init_called = 0;
static int (*orig_thr_create)(void *, size_t, void *(*)(void *), void *, long,  thread_t *);
static int (*orig_pthread_create)(pthread_t *, const pthread_attr_t *, void *(*)(void *), void *);

static const char pwd_env_prefix[] = "FSREWRITE_PWD=";

static const fptrname_t fptrname_tbl[] = {
  // Used to detect if guest is multi-threading
  {"_thr_create",     (const void **)&orig_thr_create, REQUIRED},
  {"_pthread_create", (const void **)&orig_pthread_create, REQUIRED},
  // Underlying system calls
  {"_access",      (const void **)&libc_access, REQUIRED},
  {"_acl",         (const void **)&libc_acl, REQUIRED},
  {"_chdir",       (const void **)&libc_chdir, REQUIRED},
  {"_fchdir",      (const void **)&libc_fchdir, REQUIRED},
  {"_chmod",       (const void **)&libc_chmod, REQUIRED},
  {"_chown",       (const void **)&libc_chown, REQUIRED},
  {"_lchown",      (const void **)&libc_lchown, REQUIRED},
  {"_close",       (const void **)&libc_close, REQUIRED},
  {"_dup",         (const void **)&libc_dup, REQUIRED},
  {"_execve",      (const void **)&libc_execve, REQUIRED},
  {"_fcntl",       (const void **)&libc_fcntl, REQUIRED},
  {"_getcwd",      (const void **)&libc_getcwd, REQUIRED},
  {"_getdents",    (const void **)&libc_getdents, REQUIRED},
  {"_link",        (const void **)&libc_link, REQUIRED},
  {"_mkdir",       (const void **)&libc_mkdir, REQUIRED},
  {"_mknod",       (const void **)&libc_mknod, REQUIRED},
  {"_open",        (const void **)&libc_open, REQUIRED},
  {"_openat",      (const void **)&libc_openat, OPTIONAL},
  {"_pathconf",    (const void **)&libc_pathconf, REQUIRED},
  {"_readlink",    (const void **)&libc_readlink, REQUIRED},
  {"_rename",      (const void **)&libc_rename, REQUIRED},
  {"_renameat",    (const void **)&libc_renameat, OPTIONAL},
  {"_resolvepath", (const void **)&libc_resolvepath, REQUIRED},
  {"_rmdir",       (const void **)&libc_rmdir, REQUIRED},
  {"_stat",        (const void **)&libc_stat, REQUIRED},
  {"_lstat",       (const void **)&libc_lstat, REQUIRED},
  {"_statvfs",     (const void **)&libc_statvfs, REQUIRED},
  {"_symlink",     (const void **)&libc_symlink, REQUIRED},
  {"_unlink",      (const void **)&libc_unlink, REQUIRED},
  {"_unlinkat",    (const void **)&libc_unlinkat, OPTIONAL},
  {"_utimes",      (const void **)&libc_utimes, REQUIRED},
  {"_futimesat",   (const void **)&libc_futimesat, OPTIONAL}
#ifdef __i386 // See <sys/stat_impl.h>
  ,
  {"_xmknod", (const void **)&libc_xmknod, REQUIRED},
  {"_xstat",  (const void **)&libc_xstat, REQUIRED},
  {"_lxstat", (const void **)&libc_lxstat, REQUIRED}
#endif
#ifdef _ILP32 // See lf64(5)
  ,
  {"_getdents64", (const void **)&libc_getdents64, REQUIRED},
  {"_open64",     (const void **)&libc_open64, REQUIRED},
  {"_openat64",   (const void **)&libc_openat64, OPTIONAL},
  {"_stat64",     (const void **)&libc_stat64, REQUIRED},
  {"_lstat64",    (const void **)&libc_lstat64, REQUIRED},
  {"_statvfs64",  (const void **)&libc_statvfs64, REQUIRED}
#endif
};

static void init()
{
  int i;
  void *fptr;
  
  // Read the FSREWRITE_RULES environment variable
  rules_env = getenv("FSREWRITE_RULES");
  if (!rules_env) {
    fprintf(stderr, "fsrewrite: FSREWRITE_RULES is not set.\n");
    abort();
  }
  
  // Read the FSREWRITE_PWD environment variable
  pwd_env = getenv("FSREWRITE_PWD");
  if (!pwd_env) {
    fprintf(stderr, "fsrewrite: FSREWRITE_PWD is not set.\n");
    abort();
  }
  if (strlen(pwd_env) > PATH_MAX) {
    fprintf(stderr, "fsrewrite: FSREWRITE_PWD is too long.\n");
    abort();
  }
  
  // Initialize all the function pointers
  for (i=0; i<sizeof(fptrname_tbl)/sizeof(fptrname_t); i++) {
    fptr = dlsym(RTLD_NEXT, fptrname_tbl[i].name);
    if (!fptr) {
      if (fptrname_tbl[i].required) {
        fprintf(stderr,
          "fsrewrite: Unable to a required symbol in C shared library. (%s)\n",
          fptrname_tbl[i].name);
        abort();
      }
      else {
        fptr = (void *)&abort;
      }
    }
    *(fptrname_tbl[i].fptr) = fptr;
  }
  
  // Initialize global data in other compilation unit(s)
  tracker_init();
}

static void
#ifdef __GNUC__
__attribute__ ((constructor))
#endif
constructor()
{
  INIT();
}

#ifdef __SUNPRO_C
#pragma init(constructor)
#endif

int is_mt = 0;
char *rules_env, *pwd_env;

int (*libc_access)(const char *, int);
int (*libc_acl)(const char *, int, int, void *);
int (*libc_chdir)(const char *);
int (*libc_fchdir)(int);
int (*libc_chmod)(const char *, mode_t);
int (*libc_chown)(const char *, uid_t, gid_t);
int (*libc_lchown)(const char *, uid_t, gid_t);
int (*libc_close)(int);
int (*libc_dup)(int);
int (*libc_execve)(const char *, char *const *, char *const *);
int (*libc_fcntl)(int, int, ...);
char *(*libc_getcwd)(char *, size_t);
int (*libc_getdents)(int, void *, size_t);
int (*libc_link)(const char *, const char *);
int (*libc_mkdir)(const char *, mode_t);
int (*libc_mknod)(const char *, mode_t, dev_t);
int (*libc_open)(const char *, int, ...);
int (*libc_openat)(int, const char *, int, ...);
long (*libc_pathconf)(const char *, int);
ssize_t (*libc_readlink)(const char *, char *, size_t);
int (*libc_rename)(const char *, const char *);
int (*libc_renameat)(int, const char *, int, const char *);
int (*libc_resolvepath)(const char *, char *, size_t);
int (*libc_rmdir)(const char *);
int (*libc_stat)(const char *,  void *);
int (*libc_lstat)(const char *, void *);
int (*libc_statvfs)(const char *, void *);
int (*libc_symlink)(const char *, const char *);
int (*libc_unlink)(const char *);
int (*libc_unlinkat)(int, const char *, int);
int (*libc_utimes)(const char *, const void *);
int (*libc_futimesat)(int, const char *, const void *);

#ifdef __i386
int (*libc_xmknod)(const int, const char *, mode_t, dev_t);
int (*libc_xstat)(const int, const char *, struct stat *);
int (*libc_lxstat)(const int, const char *, struct stat *);
#endif

#ifdef _ILP32
int (*libc_getdents64)(int, void *, size_t);
int (*libc_open64)(const char *, int, ...);
int (*libc_openat64)(int, const char *, int, ...);
int (*libc_stat64)(const char *, void *);
int (*libc_lstat64)(const char *, void *);
int (*libc_statvfs64)(const char *, void *);
#endif

/* ----------
 * thr_create
 * ----------
 */

int _thr_create(
  void *stack_base,
  size_t stack_size,
  void *(*start_func)(void *),
  void *arg,
  long flags,
  thread_t *new_thread_id)
{
  INIT();
  is_mt = 1;
  
  return orig_thr_create(
    stack_base, stack_size, start_func, arg, flags, new_thread_id);
}

#pragma weak thr_create = _thr_create

/* --------------
 * pthread_create
 * --------------
 */

int _pthread_create(
  pthread_t *thread,
  const pthread_attr_t *attr,
  void *(*start_routine)(void *),
  void *arg)
{
  INIT();
  is_mt = 1;
  
  return orig_pthread_create(thread, attr, start_routine, arg);
}

#pragma weak pthread_create = _pthread_create

/* ------
 * access 
 * ------
 */

typedef struct {
  int amode;
} access_params_t;

static int access_cb(const char *path, const char *vpath, void *cb_data)
{
  access_params_t *params = (access_params_t *)cb_data;
  
  return libc_access(path, params->amode);
}

int _access(const char *path, int amode)
{
  access_params_t params;
  
  INIT();
  
  params.amode = amode;
  return tracker_rewrite(AT_FDCWD, path, &access_cb, NULL, &params);
}

#pragma weak access = _access

/* ---
 * acl
 * ---
 */

typedef struct {
  int cmd;
  int nentries;
  void *aclbufp;
} acl_params_t;

static int acl_cb(const char *path, const char *vpath, void *cb_data)
{
  acl_params_t *params = (acl_params_t *)cb_data;
  
  return libc_acl(path, params->cmd, params->nentries, params->aclbufp);
}

int _acl(const char *pathp, int cmd, int nentries, void *aclbufp)
{
  acl_params_t params;
  
  INIT();
  
  params.cmd = cmd;
  params.nentries = nentries;
  params.aclbufp = aclbufp;
  return tracker_rewrite(AT_FDCWD, pathp, &acl_cb, NULL, &params);
}

#pragma weak acl = _acl

/* -----
 * chdir
 * -----
 */

int _chdir(const char *path)
{
  INIT();
  
  return tracker_chdir(path);
}

#pragma weak chdir = _chdir

/* ------
 * fchdir
 * ------
 */

int _fchdir(int fildes)
{
  INIT();
  
  return tracker_fchdir(fildes);
}

#pragma weak fchdir = _fchdir

/* -----
 * chmod
 * -----
 */

typedef struct {
  mode_t mode;
} chmod_params_t;

static int chmod_cb(const char *path, const char *vpath, void *cb_data)
{
  chmod_params_t *params = (chmod_params_t *)cb_data;
  
  return libc_chmod(path, params->mode);
}

int _chmod(const char *path, mode_t mode)
{
  chmod_params_t params;
  
  INIT();
  
  params.mode = mode;
  return tracker_rewrite(AT_FDCWD, path, &chmod_cb, NULL, &params);
}

#pragma weak chmod = _chmod

/* -----
 * chown
 * -----
 */

typedef struct {
  uid_t owner;
  gid_t group;
} chown_params_t;

static int chown_cb(const char *path, const char *vpath, void *cb_data)
{
  chown_params_t *params = (chown_params_t *)cb_data;
  
  return libc_chown(path, params->owner, params->group);
}

int _chown(const char *path, uid_t owner, gid_t group)
{
  chown_params_t params;
  
  INIT();
  
  params.owner = owner;
  params.group = group;
  return tracker_rewrite(AT_FDCWD, path, &chown_cb, NULL, &params);
}

#pragma weak chown = _chown

/* ------
 * lchown
 * ------
 */

static int lchown_cb(const char *path, const char *vpath, void *cb_data)
{
  chown_params_t *params = (chown_params_t *)cb_data;
  
  return libc_lchown(path, params->owner, params->group);
}

int _lchown(const char *path, uid_t owner, gid_t group)
{
  chown_params_t params;
  
  INIT();
  
  params.owner = owner;
  params.group = group;
  return tracker_rewrite(AT_FDCWD, path, &lchown_cb, NULL, &params);
}

#pragma weak lchown = _lchown

/* -----
 * close
 * -----
 */

int _close(int fildes)
{
  int ret;
  
  INIT();
  
  ret = libc_close(fildes);
  if (ret >= 0) {
    tracker_close_fd(fildes);
  }
  return ret;
}

#pragma weak close = _close

/* ---
 * dup
 * ---
 */

int _dup(int fildes)
{
  int ret;
  
  INIT();
  
  ret = libc_dup(fildes);
  if (ret >= 0) {
    tracker_dup_fd(fildes, ret);
  }
  return ret;
}

#pragma weak dup = _dup

/* ------
 * execve
 * ------
 */

typedef struct {
  char *const *argv;
  char *const *envp;
} execve_params_t;

static int execve_cb(const char *path, const char *vpath, void *cb_data)
{
  execve_params_t *params = (execve_params_t *)cb_data;
  
  return libc_execve(path, params->argv, params->envp);
}

int _execve(const char *path, char *const *argv, char *const *envp)
{
  execve_params_t params;
  char cwd_buf[PATH_MAX];
  path_buf_t cwd = {cwd_buf, 0};
  int i, pwd_env_found, pwd_env_index, pwd_env_correct;
  
  INIT();
  
  tracker_getcwd(cwd.buf, PATH_MAX);
  cwd.len = strlen(cwd.buf);
  
  i = pwd_env_found = pwd_env_index = pwd_env_correct = 0;
  if (envp) {
    for (; envp[i]; i++) {
      if (strncmp(envp[i], pwd_env_prefix, strlen(pwd_env_prefix)) == 0) {
        pwd_env_found = 1; 
        pwd_env_index = i;
        if (strcmp(envp[i] + strlen(pwd_env_prefix), cwd.buf) == 0) {
          pwd_env_correct = 1;
        }
      }
    }
  }
  if (!pwd_env_correct) {
    char **new_envp, *envvar;
    
    new_envp = alloca((i + 1 + !pwd_env_found) * sizeof(char *));
    if (envp) {
      memcpy(new_envp, envp, (i + 1) * sizeof(char *));
    }
    if (!pwd_env_found) {
      new_envp[i + 1] = NULL;
      pwd_env_index = i;
    }
    envvar = alloca(strlen(pwd_env_prefix) + cwd.len + 1);
    strcpy(envvar, pwd_env_prefix);
    strcpy(envvar + strlen(pwd_env_prefix), cwd.buf);
    new_envp[pwd_env_index] = envvar;
    envp = (char *const *)new_envp;
  }
  
  params.argv = argv;
  params.envp = envp;
  return tracker_rewrite(AT_FDCWD, path, &execve_cb, NULL, &params);
}

#pragma weak execve = _execve

/* -----
 * fcntl
 * -----
 */

int _fcntl(int fildes, int cmd, ...)
{
  va_list vargs;

  INIT();
  
  va_start(vargs, cmd);
  
  switch (cmd) {
    case F_DUPFD: {
      int ret, arg = va_arg(vargs, int);
      va_end(vargs);
      ret = libc_fcntl(fildes, cmd, arg);
      if (ret >= 0) {
        tracker_dup_fd(fildes, ret);
      }
      return ret;
    }
    
    case F_DUP2FD: {
      int ret, arg = va_arg(vargs, int);
      va_end(vargs);
      ret = libc_fcntl(fildes, cmd, arg);
      if (ret >= 0) {
        tracker_close_fd(ret);
        tracker_dup_fd(fildes, ret);
      }
      return ret;
    }
    
    case F_GETFD: // arg is not specified
    case F_GETFL:
    case F_GETOWN:
    case F_GETXFL: {
      va_end(vargs);
      return libc_fcntl(fildes, cmd);
    }
    
    case F_SETFD: // arg is of type int
    case F_SETFL:
    case F_SETOWN: {
      int arg = va_arg(vargs, int);
      va_end(vargs);
      return libc_fcntl(fildes, cmd, arg);
    }   
    
    case F_FREESP: { // arg is of type struct flock
      struct flock arg = va_arg(vargs, struct flock);
      va_end(vargs);
      return libc_fcntl(fildes, cmd, arg);
    }
    
#ifdef _ILP32
    case F_FREESP64: { // arg is of type struct flock64
      struct flock64 arg = va_arg(vargs, struct flock64);
      va_end(vargs);
      return libc_fcntl(fildes, cmd, arg);
    }
#endif
    
    case F_GETLK: // arg is of type void*
    case F_SETLK:
    case F_SETLKW:
#ifdef _ILP32
    case F_GETLK64:
    case F_SETLK64:
    case F_SETLKW64:
#endif
    case F_SHARE:
    case F_UNSHARE: {
      void *arg = va_arg(vargs, void *);
      va_end(vargs);
      return libc_fcntl(fildes, cmd, arg);
    }
    
    default: // unknown command
      va_end(vargs);
      errno = EINVAL;
      return -1;
  }
}

#pragma weak fcntl = _fcntl

/* ------
 * getcwd
 * ------
 */

char *_getcwd(char *buf, size_t size)
{
  INIT();
  
  return tracker_getcwd(buf, size);
}

#pragma weak getcwd = _getcwd

/* --------
 * getdents
 * --------
 */

int _getdents(int fildes, struct dirent *buf, size_t nbyte)
{
  INIT();
  
  return tracker_getdents(fildes, buf, nbyte);
}

#pragma weak getdents = _getdents

/* ----
 * link
 * ----
 */

typedef struct {
  const char *existing_vpath;
  const char *new_path;
} link_params_t;

static int link2_cb(const char *path, const char *vpath, void *cb_data)
{
  link_params_t *params = (link_params_t *)cb_data;
  
  return libc_link(path, params->new_path);
}

static int link_cb(const char *path, const char *vpath, void *cb_data)
{
  link_params_t *params = (link_params_t *)cb_data;
  
  params->new_path = path;
  return tracker_rewrite(
    AT_FDCWD, params->existing_vpath, &link2_cb, NULL, params);
}

int _link(const char *existing, const char *new)
{
  link_params_t params;
  
  INIT();
  
  params.existing_vpath = existing;
  return tracker_rewrite(AT_FDCWD, new, &link_cb, NULL, &params);
}

#pragma weak link = _link

/* -----
 * mkdir
 * -----
 */

typedef struct {
  mode_t mode;
} mkdir_params_t;

static int mkdir_cb(const char *path, const char *vpath, void *cb_data)
{
  mkdir_params_t *params = (mkdir_params_t *)cb_data;
  
  return libc_mkdir(path, params->mode);
}

int _mkdir(const char *path, mode_t mode)
{
  mkdir_params_t params;
  
  INIT();
  
  params.mode = mode;
  return tracker_rewrite(AT_FDCWD, path, &mkdir_cb, NULL, &params);
}

#pragma weak mkdir = _mkdir

/* -----
 * mknod
 * -----
 */

typedef struct {
  mode_t mode;
  dev_t dev;
} mknod_params_t;

static int mknod_cb(const char *path, const char *vpath, void *cb_data)
{
  mknod_params_t *params = (mknod_params_t *)cb_data;
  
  return libc_mknod(path, params->mode, params->dev);
}

int _mknod(const char *path, mode_t mode, dev_t dev)
{
  mknod_params_t params;
  
  INIT();
  
  params.mode = mode;
  params.dev = dev;
  return tracker_rewrite(AT_FDCWD, path, &mknod_cb, NULL, &params);
}

#pragma weak mknod = _mknod

/* ----
 * open
 * ----
 */

typedef struct {
  int oflag;
  int has_mode;
  mode_t mode;
} open_params_t;

static int open_cb(const char *path, const char *vpath, void *cb_data)
{
  open_params_t *params = (open_params_t *)cb_data;
  int fd;
  
  if (params->has_mode) {
    fd = libc_open(path, params->oflag, params->mode);
  }
  else {
    fd = libc_open(path, params->oflag);
  }
  if (fd >= 0) {
    if (tracker_open_fd(fd, vpath) < 0) {
      close(fd);
      return -1;
    }
  }
  return fd;
}

int _open(const char *path, int oflag, ...)
{
  open_params_t params;
  mode_t mode;
  int retval;
  
  INIT();
  
  if (oflag & O_CREAT) {
    va_list vargs;
    
    va_start(vargs, oflag);
    mode = va_arg(vargs, mode_t);
    va_end(vargs);
  }
  
  if (oflag & O_XATTR) {
    return libc_open(path, oflag, mode);
  }
  
  params.oflag = oflag;
  params.has_mode = 0;
  if (oflag & O_CREAT) {
    params.oflag &= ~O_CREAT;
    retval = tracker_rewrite(AT_FDCWD, path, &open_cb, NULL, &params);
    if (retval < 0) {
      params.oflag = oflag;
      params.has_mode = 1;
      params.mode = mode;
      return tracker_rewrite(AT_FDCWD, path, &open_cb, NULL, &params);
    }
    return retval;
  }
  else {
    return tracker_rewrite(AT_FDCWD, path, &open_cb, NULL, &params);
  }
}

#pragma weak open = _open

/* ------
 * openat
 * ------
 */

static int openat_failed_cb(int fd, const char *path, void *cb_data)
{
  open_params_t *params = (open_params_t *)cb_data;
  
  if (params->has_mode) {
    return libc_openat(fd, path, params->oflag, params->mode);
  }
  else {
    return libc_openat(fd, path, params->oflag);
  }
}

int _openat(int fildes, const char *path, int oflag, ...)
{
  open_params_t params;
  mode_t mode;
  int retval;
  
  INIT();
  
  if (oflag & O_CREAT) {
    va_list vargs;
    
    va_start(vargs, oflag);
    mode = va_arg(vargs, mode_t);
    va_end(vargs);
  }
  
  if (oflag & O_XATTR) {
    return libc_openat(fildes, path, oflag, mode);
  }
  
  params.oflag = oflag;
  params.has_mode = 0;
  if (oflag & O_CREAT) {
    params.oflag &= ~O_CREAT;
    retval = tracker_rewrite(
      fildes, path, &open_cb, &openat_failed_cb, &params);
    if (retval < 0) {
      params.oflag = oflag;
      params.has_mode = 1;
      params.mode = mode;
      return tracker_rewrite(
        fildes, path, &open_cb, &openat_failed_cb, &params);
    }
    return retval;
  }
  else {
    return tracker_rewrite(
      fildes, path, &open_cb, &openat_failed_cb, &params);
  }
}

#pragma weak openat = _openat

/* -----
 * creat
 * -----
 */

int _creat(const char *path, mode_t mode)
{
  INIT();
  
  return open(path, O_WRONLY | O_CREAT | O_TRUNC, mode);
}

#pragma weak creat = _creat

/* --------
 * pathconf
 * --------
 */

typedef struct {
  int name;
  long lret;
} pathconf_params_t;

static int pathconf_cb(const char *path, const char *vpath, void *cb_data)
{
  pathconf_params_t *params = (pathconf_params_t *)cb_data;
  
  params->lret = libc_pathconf(path, params->name);
  return (params->lret < 0) ? -1 : 0;
}

long _pathconf(const char *path, int name)
{
  pathconf_params_t params;
  
  INIT();
  
  params.name = name;
  tracker_rewrite(AT_FDCWD, path, &pathconf_cb, NULL, &params);
  return params.lret;
}

#pragma weak pathconf = _pathconf

/* --------
 * readlink
 * --------
 */

typedef struct {
  char *buf;
  size_t bufsiz;
  ssize_t lret;
} readlink_params_t;

static int readlink_cb(const char *path, const char *vpath, void *cb_data)
{
  readlink_params_t *params = (readlink_params_t *)cb_data;

  params->lret = libc_readlink(path, params->buf, params->bufsiz);
  return (params->lret < 0) ? -1 : 0;
}

ssize_t _readlink(const char *path, char *buf, size_t bufsiz)
{
  readlink_params_t params;
  
  INIT();
  
  params.buf = buf;
  params.bufsiz = bufsiz;
  tracker_rewrite(AT_FDCWD, path, &readlink_cb, NULL, &params);
  return params.lret;
}

#pragma weak readlink = _readlink

/* ------
 * rename
 * ------
 */

typedef struct {
  const char *new_vpath;
  const char *old_path;
} rename_params_t;

static int rename2_cb(const char *path, const char *vpath, void *cb_data)
{
  rename_params_t *params = (rename_params_t *)cb_data;
  
  return libc_rename(params->old_path, path);
}

static int rename_cb(const char *path, const char *vpath, void *cb_data)
{
  rename_params_t *params = (rename_params_t *)cb_data;
  
  params->old_path = path;
  return tracker_rewrite(AT_FDCWD, params->new_vpath, &rename2_cb, NULL, params);
}

int _rename(const char *old, const char *new)
{
  rename_params_t params;
  
  INIT();
  
  params.new_vpath = new;
  return tracker_rewrite(AT_FDCWD, old, &rename_cb, NULL, &params);
}

#pragma weak rename = _rename

/* --------
 * renameat
 * --------
 */

typedef struct {
  int fromfd;
  int tofd;
  const char *old_vpath;
  const char *new_vpath;
  const char *old_path;
} renameat_params_t;

static int renameat_rewrite2_cb(
  const char *path, const char *vpath, void *cb_data)
{
  renameat_params_t *params = (renameat_params_t *)cb_data;
  
  if (params->old_path) {
    return libc_rename(params->old_path, path);
  }
  else {
    return libc_renameat(params->fromfd, params->old_vpath, -1, path);
  }
}

static int renameat_unknownfd2_cb(int fd, const char *path, void *cb_data)
{
  renameat_params_t *params = (renameat_params_t *)cb_data;
  
  if (params->old_path) {
    return libc_renameat(-1, params->old_path, fd, path);
  }
  else {
    return libc_renameat(params->fromfd, params->old_vpath, fd, path);
  }
}

static int renameat_rewrite_cb(
  const char *path, const char *vpath, void *cb_data)
{
  renameat_params_t *params = (renameat_params_t *)cb_data;
  
  params->old_path = path;
  return tracker_rewrite(
    params->tofd, params->new_vpath,
    &renameat_rewrite2_cb, &renameat_unknownfd2_cb, params);
}

static int renameat_unknownfd_cb(int fd, const char *path, void *cb_data)
{
  renameat_params_t *params = (renameat_params_t *)cb_data;
  
  params->old_path = NULL;
  return tracker_rewrite(
    params->tofd, params->new_vpath,
    &renameat_rewrite2_cb, &renameat_unknownfd2_cb, params);
}

int _renameat(int fromfd, const char *old, int tofd, const char *new)
{
  renameat_params_t params;
  
  INIT();
  
  params.fromfd = fromfd;
  params.tofd = tofd;
  params.old_vpath = old;
  params.new_vpath = new;
  return tracker_rewrite(
    fromfd, old, &renameat_rewrite_cb, &renameat_unknownfd_cb, &params);
}

#pragma weak renameat = _renameat

/* -----------
 * resolvepath
 * -----------
 */

typedef struct {
  char *buf;
  size_t bufsize;
} resolvepath_params_t;

static int resolvepath_cb(const char *path, const char *vpath, void *cb_data)
{
  resolvepath_params_t *params = (resolvepath_params_t *)cb_data;
  
  return libc_resolvepath(path, params->buf, params->bufsize);
}

int _resolvepath(const char *path, char *buf, size_t bufsize)
{
  resolvepath_params_t params;
  
  params.buf = buf;
  params.bufsize = bufsize;
  return tracker_rewrite(AT_FDCWD, path, &resolvepath_cb, NULL, &params);
}

#pragma weak resolvepath = _resolvepath

/* -----
 * rmdir
 * -----
 */

static int rmdir_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_rmdir(path);
}

int _rmdir(const char *path)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &rmdir_cb, NULL, NULL);
}

#pragma weak rmdir = _rmdir

/* ----
 * stat
 * ----
 */

static int stat_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_stat(path, (struct stat *)cb_data);
}

int _stat(const char *path, struct stat *buf)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &stat_cb, NULL, buf);
}

#pragma weak stat = _stat

/* -----
 * lstat
 * -----
 */

static int lstat_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_lstat(path, (struct stat *)cb_data);
}

int _lstat(const char *path, struct stat *buf)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &lstat_cb, NULL, buf);
}

#pragma weak lstat = _lstat

/* -------
 * statvfs
 * -------
 */

static int statvfs_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_statvfs(path, (struct statvfs *)cb_data);
}

int _statvfs(const char *path, struct statvfs *buf)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &statvfs_cb, NULL, buf);
}

#pragma weak statvfs = _statvfs

/* -------
 * symlink
 * -------
 */

typedef struct {
  const char *existing;
} symlink_params_t;

static int symlink_cb(const char *path, const char *vpath, void *cb_data)
{
  symlink_params_t *params = (symlink_params_t *)cb_data;
  
  return libc_symlink(params->existing, path);
}

int _symlink(const char *existing, const char *new)
{
  symlink_params_t params;
  
  INIT();
  
  params.existing = existing;
  return tracker_rewrite(AT_FDCWD, new, &symlink_cb, NULL, &params);
}

#pragma weak symlink = _symlink

/* ------
 * unlink
 * ------
 */

static int unlink_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_unlink(path);
}

int _unlink(const char *path)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &unlink_cb, NULL, NULL);
}

#pragma weak unlink = _unlink

/* --------
 * unlinkat
 * --------
 */

typedef struct {
  int flag;
} unlinkat_params_t;

static int unlinkat_path_rewrite_cb(
  const char *path, const char *vpath, void *cb_data)
{
  unlinkat_params_t *params = (unlinkat_params_t *)cb_data;
  
  if (params->flag == AT_REMOVEDIR) {
    return libc_rmdir(path);
  }
  else {
    return libc_unlink(path);
  }
}

static int unlinkat_failed_cb(int fd, const char *path, void *cb_data)
{
  unlinkat_params_t *params = (unlinkat_params_t *)cb_data;
  
  return libc_unlinkat(fd, path, params->flag);
}

int _unlinkat(int dirfd, const char *path, int flag)
{
  unlinkat_params_t params;
  
  INIT();
  
  params.flag = flag;
  return tracker_rewrite(
    dirfd, path, &unlinkat_path_rewrite_cb, &unlinkat_failed_cb, &params);
}

#pragma weak unlinkat = _unlinkat

/* ------
 * utimes
 * ------
 */

static int utimes_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_utimes(path, (const struct timeval *)cb_data);
}

int _utimes(const char *path, const struct timeval times[2])
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &utimes_cb, NULL, (void *)times);
}

#pragma weak utimes = _utimes

/* ---------
 * futimesat
 * ---------
 */

static int futimesat_failed_cb(int fd, const char *path, void *cb_data)
{
  return libc_futimesat(fd, path, (const struct timeval *)cb_data);
}

int _futimesat(int fildes, const char *path, const struct timeval times[2])
{
  INIT();
  
  return tracker_rewrite(
    fildes, path, &utimes_cb, &futimesat_failed_cb, (void *)times);
}

#pragma weak futimesat = _futimesat

#ifdef __i386 // See <sys/stat_impl.h>

/* ------
 * xmknod
 * ------
 */

typedef struct {
  int ver;
  mode_t mode;
  dev_t dev;
} xmknod_params_t;

static int xmknod_cb(const char *path, const char *vpath, void *cb_data)
{
  xmknod_params_t *params = (xmknod_params_t *)cb_data;
  
  return libc_xmknod(params->ver, path, params->mode, params->dev);
}

int _xmknod(const int ver, const char *path, mode_t mode, dev_t dev)
{
  xmknod_params_t params;
  
  INIT();
  
  params.ver = ver;
  params.mode = mode;
  params.dev = dev;
  return tracker_rewrite(AT_FDCWD, path, &xmknod_cb, NULL, &params);
}

/* -----
 * xstat
 * -----
 */

typedef struct {
  int ver;
  struct stat *buf;
} xstat_params_t;

static int xstat_cb(const char *path, const char *vpath, void *cb_data)
{
  xstat_params_t *params = (xstat_params_t *)cb_data;
  
  return libc_xstat(params->ver, path, params->buf);
}

int _xstat(const int ver, const char *path, struct stat *buf)
{
  xstat_params_t params;
  
  INIT();
  
  params.ver = ver;
  params.buf = buf;
  return tracker_rewrite(AT_FDCWD, path, &xstat_cb, NULL, &params);
}

/* ------
 * lxstat
 * ------
 */

static int lxstat_cb(const char *path, const char *vpath, void *cb_data)
{
  xstat_params_t *params = (xstat_params_t *)cb_data;
  
  return libc_lxstat(params->ver, path, params->buf);
}

int _lxstat(const int ver, const char *path, struct stat *buf)
{
  xstat_params_t params;
  
  INIT();
  
  params.ver = ver;
  params.buf = buf;
  return tracker_rewrite(AT_FDCWD, path, &lxstat_cb, NULL, &params);
}

#endif

#ifdef _ILP32 // See lf64(5)

/* ----------
 * getdents64
 * ----------
 */

int _getdents64(int fildes, struct dirent *buf, size_t nbyte)
{
  INIT();
  
  return tracker_getdents64(fildes, buf, nbyte);
}

#pragma weak getdents64 = _getdents64

/* ------
 * open64
 * ------
 */

static int open64_cb(const char *path, const char *vpath, void *cb_data)
{
  open_params_t *params = (open_params_t *)cb_data;
  int fd;
  
  fd = libc_open64(path, params->oflag, params->mode);
  if (fd >= 0) {
    if (tracker_open_fd(fd, vpath) < 0) {
      close(fd);
      return -1;
    }
  }
  return fd;
}

int _open64(const char *path, int oflag, ...)
{
  open_params_t params;
  mode_t mode;
  int retval;
  
  INIT();
  
  if (oflag & O_CREAT) {
    va_list vargs;
    
    va_start(vargs, oflag);
    mode = va_arg(vargs, mode_t);
    va_end(vargs);
  }
  
  if (oflag & O_XATTR) {
    return libc_open64(path, oflag, mode);
  }
  
  params.oflag = oflag;
  params.has_mode = 0;
  if (oflag & O_CREAT) {
    params.oflag &= ~O_CREAT;
    retval = tracker_rewrite(AT_FDCWD, path, &open64_cb, NULL, &params);
    if (retval < 0) {
      params.oflag = oflag;
      params.has_mode = 1;
      params.mode = mode;
      return tracker_rewrite(AT_FDCWD, path, &open64_cb, NULL, &params);
    }
    return retval;
  }
  else {
    return tracker_rewrite(AT_FDCWD, path, &open64_cb, NULL, &params);
  }
}

#pragma weak open64 = _open64

/* --------
 * openat64
 * --------
 */

static int openat64_failed_cb(int fd, const char *path, void *cb_data)
{
  open_params_t *params = (open_params_t *)cb_data;
  
  if (params->has_mode) {
    return libc_openat64(fd, path, params->oflag, params->mode);
  }
  else {
    return libc_openat64(fd, path, params->oflag);
  }
}

int _openat64(int fildes, const char *path, int oflag, ...)
{
  open_params_t params;
  mode_t mode;
  int retval;
  
  INIT();
  
  if (oflag & O_CREAT) {
    va_list vargs;
    
    va_start(vargs, oflag);
    mode = va_arg(vargs, mode_t);
    va_end(vargs);
  }
  
  if (oflag & O_XATTR) {
    return libc_openat64(fildes, path, oflag, mode);
  }

  params.oflag = oflag;
  params.has_mode = 0;
  if (oflag & O_CREAT) {
    params.oflag &= ~O_CREAT;
    retval = tracker_rewrite(AT_FDCWD, path, &open64_cb, &openat64_failed_cb, &params);
    if (retval < 0) {
      params.oflag = oflag;
      params.has_mode = 1;
      params.mode = mode;
      return tracker_rewrite(AT_FDCWD, path, &open64_cb, &openat64_failed_cb, &params);
    }
    return retval;
  }
  else {
    return tracker_rewrite(AT_FDCWD, path, &open64_cb, &openat64_failed_cb, &params);
  }
}

#pragma weak openat64 = _openat64

/* -------
 * creat64
 * -------
 */

int _creat64(const char *path, mode_t mode)
{
  INIT();
  
  return _open64(path, O_WRONLY | O_CREAT | O_TRUNC, mode);
}

#pragma weak creat64 = _creat64

/* ------
 * stat64
 * ------
 */

static int stat64_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_stat64(path, (struct stat64 *)cb_data);
}

int _stat64(const char *path, struct stat *buf)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &stat64_cb, NULL, buf);
}

#pragma weak stat64 = _stat64

/* -------
 * lstat64
 * -------
 */

static int lstat64_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_lstat64(path, (struct stat64 *)cb_data);
}

int _lstat64(const char *path, struct stat *buf)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &lstat64_cb, NULL, buf);
}

#pragma weak lstat64 = _lstat64

/* ---------
 * statvfs64
 * ---------
 */

static int statvfs64_cb(const char *path, const char *vpath, void *cb_data)
{
  return libc_statvfs64(path, (struct statvfs64 *)cb_data);
}

int _statvfs64(const char *path, struct statvfs *buf)
{
  INIT();
  
  return tracker_rewrite(AT_FDCWD, path, &statvfs64_cb, NULL, buf);
}

#pragma weak statvfs64 = _statvfs64

#endif
