/*
 * tracker.c
 * Copyright (C) 2005 Robin KAY. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 */

#include "compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>

#include "utils.h"
#include "mallory.h"
#include "tracker.h"

typedef enum {
  NOT_TRACKED = 0,
  DIRECTORY,
  DUPLICATE
} rec_type_t;

typedef struct {
  rec_type_t rec_type;
  int next_dup;
  union {
    struct {
      char vpath_buf[PATH_MAX];
      int vpath_len;
      tree_node_t *dirents_root;
      tree_node_t *dirents_first;
      tree_node_t *dirents_curr;
    } dir;
    struct {
      int prev_dup;
    } dup;
  } rec;
} fdinfo_t;

#define GETDENTS_BUFSIZE 4096

static pthread_mutex_t big_giant_lock = PTHREAD_MUTEX_INITIALIZER;
static char cwd_buf[PATH_MAX];
static path_buf_t cwd = {cwd_buf, 0};
static int page_size, sub_page_size, sub_pages_per_page;
static int fdinfo_tbl_num_pages, fdinfo_tbl_num_sub_pages;
static void *fdinfo_tbl = NULL;

static fdinfo_t *get_fd_record(int fd)
{
  return (fdinfo_t *)((char *)fdinfo_tbl + fd * sub_page_size);
}

static int traverse_dup_chain(int fd)
{
  fdinfo_t *fdinfo;
  
  // Find the first record in any duplicate chain
  fdinfo = get_fd_record(fd);
  while (fdinfo->rec_type == DUPLICATE) {
    fd = fdinfo->rec.dup.prev_dup;
    fdinfo = get_fd_record(fd);
  }
  
  return fd;
}

int get_path(path_buf_t *wpath, int fd)
{
  fdinfo_t *fdinfo;
  
  // Return the current working directory if AT_FDCWD is passed
  if (fd == AT_FDCWD) {
    MUTEX_LOCK(&big_giant_lock);
    path_append(path_clear(wpath), cwd.buf, cwd.len);
    MUTEX_UNLOCK(&big_giant_lock);
    return 0;
  }
  
  // Return an error if the file descriptor is outside the table
  if ((fd < 0) || (fd > fdinfo_tbl_num_sub_pages)) {
    errno = EBADF;
    return -1;
  }
  
  // Find the start of any duplicate chain
  MUTEX_LOCK(&big_giant_lock);
  fd = traverse_dup_chain(fd);
  
  // Get the path from the record
  fdinfo = get_fd_record(fd);
  switch (fdinfo->rec_type) {
    case DIRECTORY: {
      memcpy(wpath->buf, fdinfo->rec.dir.vpath_buf, fdinfo->rec.dir.vpath_len);
      wpath->len = fdinfo->rec.dir.vpath_len;
      MUTEX_UNLOCK(&big_giant_lock);
      return 0;
    }
    
    default: {
      MUTEX_UNLOCK(&big_giant_lock);
      errno = ENOTDIR;
      return -1;
    }
  }
}

void tracker_init()
{
  struct rlimit rlim;
  
  // Copy pwd_env into cwd buffer
  path_append(&cwd, pwd_env, strlen(pwd_env));
  
  // Calculate how many records can be fit into a page
  page_size = sub_page_size = getpagesize();
  if (sizeof(fdinfo_t) > getpagesize()) {
    fprintf(stderr, "fsrewrite: Page size is too small.\n");
    abort();
  }
  while ((sub_page_size / 2) > sizeof(fdinfo_t)) {
    sub_page_size /= 2;
  }
  sub_pages_per_page = page_size / sub_page_size;
  
  // Allocate a table to hold records on file descriptors up to the hard limit
  getrlimit(RLIMIT_NOFILE, &rlim);
  fdinfo_tbl_num_pages = (rlim.rlim_max / sub_pages_per_page) + 1;
  fdinfo_tbl_num_sub_pages = fdinfo_tbl_num_pages * sub_pages_per_page;
  fdinfo_tbl = mmap(
    NULL, page_size * fdinfo_tbl_num_pages,
    PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, -1, 0);
  if (fdinfo_tbl == MAP_FAILED) {
    fprintf(
      stderr,
      "fsrewrite: Insufficient free address space for descriptor table.\n");
    abort();
  }
}

int tracker_open_fd(int fd, const char *vpath)
{
  struct stat statbuf;
  int ret;
  fdinfo_t *fdinfo;
  
  // Return an error if the file descriptor is outside the table
  if ((fd < 0) || (fd > fdinfo_tbl_num_sub_pages)) {
    errno = EBADF;
    return -1;
  }
  
  // Check that the file descriptor refers to a directory
  while (((ret = fstat(fd, &statbuf)) < 0) && (errno == EINTR)) {}
  if (ret < 0) {
    return -1;
  }
  if ((statbuf.st_mode & S_IFMT) != S_IFDIR) {
    return 0;
  }
  
  // Initialize new directory record
  MUTEX_LOCK(&big_giant_lock);
  fdinfo = get_fd_record(fd);
  fdinfo->rec_type = DIRECTORY;
  fdinfo->next_dup = -1;
  fdinfo->rec.dir.vpath_len = strlen(vpath);
  memcpy(fdinfo->rec.dir.vpath_buf, vpath, fdinfo->rec.dir.vpath_len);
  fdinfo->rec.dir.dirents_root = NULL;
  fdinfo->rec.dir.dirents_first = NULL;
  fdinfo->rec.dir.dirents_curr = NULL;
  MUTEX_UNLOCK(&big_giant_lock);
  
  return 0;
}

int tracker_dup_fd(int fd, int fd2)
{
  fdinfo_t *fdinfo;
  int last_fd;
  
  // Return an error if the file descriptor is outside the table
  if (((fd < 0) || (fd > fdinfo_tbl_num_sub_pages)) ||
      ((fd2 < 0) || (fd2 > fdinfo_tbl_num_sub_pages))) {
    errno = EBADF;
    return -1;
  }
  
  // Check that the source file descriptor is being tracked
  MUTEX_LOCK(&big_giant_lock);
  fdinfo = get_fd_record(fd);
  if (fdinfo->rec_type == NOT_TRACKED) {
    MUTEX_UNLOCK(&big_giant_lock);
    return 0;
  }
  
  // Find the last record in the duplicate chain
  last_fd = fd;
  while (fdinfo->next_dup >= 0) {
    last_fd = fdinfo->next_dup;
    fdinfo = get_fd_record(last_fd);
  }
  
  // Initialize new duplicate record
  fdinfo->next_dup = fd2;
  fdinfo = get_fd_record(fd2);
  fdinfo->rec_type = DUPLICATE;
  fdinfo->next_dup = -1;
  fdinfo->rec.dup.prev_dup = last_fd;
  MUTEX_UNLOCK(&big_giant_lock);
  
  return 0;
}

int tracker_close_fd(int fd)
{
  fdinfo_t *fdinfo;
  int first_sub_page, page_empty, i;
  
  // Return an error if the file descriptor is outside the table
  if ((fd < 0) || (fd > fdinfo_tbl_num_sub_pages)) {
    errno = EBADF;
    return -1;
  }
  
  // Check that the source file descriptor is being tracked
  MUTEX_LOCK(&big_giant_lock);
  fdinfo = get_fd_record(fd);
  if (fdinfo->rec_type == NOT_TRACKED) {
    MUTEX_UNLOCK(&big_giant_lock);
    return 0;
  }

  // Remove the record from any duplicate chain
  if (fdinfo->rec_type == DUPLICATE) {
    fdinfo_t *prev_dup;
    
    prev_dup = get_fd_record(fdinfo->rec.dup.prev_dup);
    prev_dup->next_dup = fdinfo->next_dup;
    
    if (fdinfo->next_dup >= 0) {
      fdinfo_t *next_dup;
      
      next_dup = get_fd_record(fdinfo->next_dup);
      next_dup->rec.dup.prev_dup = fdinfo->rec.dup.prev_dup;
    }
  }
  else if (fdinfo->next_dup >= 0) {
    fdinfo_t *next_dup;
    int after_next_dup;
    
    next_dup = get_fd_record(fdinfo->next_dup);
    after_next_dup = next_dup->next_dup;
    memcpy(next_dup, fdinfo, sizeof(fdinfo_t));
    next_dup->next_dup = after_next_dup;
  }
  
  // Free dirents if record is a directory
  if (fdinfo->rec_type == DIRECTORY) {
    tree_cleanup(&fdinfo->rec.dir.dirents_root);
  }
  
  // Free the sub page
  fdinfo->rec_type = NOT_TRACKED;
  first_sub_page = fd - (fd % sub_pages_per_page);
  page_empty = 1;
  for (i=first_sub_page; i<first_sub_page+sub_pages_per_page; i++) {
    fdinfo = get_fd_record(i);
    if (fdinfo->rec_type != NOT_TRACKED) {
      page_empty = 0;
      break;
    }
  }
  
  // Mark the page as free if all sub pages are free
  if (page_empty) {
    madvise(
      (char *)fdinfo_tbl + first_sub_page * sub_page_size, page_size, MADV_FREE);
  }
  MUTEX_UNLOCK(&big_giant_lock);
  
  return 0;
}

int tracker_rewrite(
  int fd,
  const char *path,
  path_rewrite_cb_t rewrite_cb,
  tracker_failed_cb_t failed_cb,
  void *cb_data)
{
  char wpath_buf[PATH_MAX];
  path_buf_t wpath = {wpath_buf, 0};
  
  // Get path to file descriptor
  MUTEX_LOCK(&big_giant_lock);
  if (get_path(&wpath, fd) < 0) {
    MUTEX_UNLOCK(&big_giant_lock);
    if (failed_cb) {
      return failed_cb(fd, path, cb_data);
    }
    else {
      errno = ENOTDIR;
      return -1;
    }
  }
  
  // Convert relative paths to absolute form
  if (!path_absolute(&wpath, path)) {
    MUTEX_UNLOCK(&big_giant_lock);
    return -1;
  }
  MUTEX_UNLOCK(&big_giant_lock);
  path_terminate(&wpath);
  
  return path_rewrite(wpath.buf, rewrite_cb, cb_data);
}

/* -----
 * chdir
 * -----
 */

static int chdir_cb(const char *path, const char *vpath, void *cb_data)
{
  int retval;
  
  retval = libc_chdir(path);
  if (retval >= 0) {
    path_append(path_clear(&cwd), vpath, strlen(vpath));
  }
  return retval;
}

int tracker_chdir(const char *path)
{
  char wpath_buf[PATH_MAX];
  path_buf_t wpath = {wpath_buf, 0};
  int retval;
  
  MUTEX_LOCK(&big_giant_lock);
  if (!path_absolute(path_append(&wpath, cwd.buf, cwd.len), path)) {
    MUTEX_UNLOCK(&big_giant_lock);
    return -1;
  }
  path_terminate(&wpath);
  retval = path_rewrite(wpath.buf, &chdir_cb, NULL);
  MUTEX_UNLOCK(&big_giant_lock);
  
  return retval;
}

/* ------
 * fchdir
 * ------
 */

int tracker_fchdir(int fd)
{
  char wpath_buf[PATH_MAX];
  path_buf_t wpath = {wpath_buf, 0};
  int retval;
  
  MUTEX_LOCK(&big_giant_lock);
  if (get_path(&wpath, fd) < 0) {
    MUTEX_UNLOCK(&big_giant_lock);
    return -1;
  }
  path_terminate(&wpath);
  retval = path_rewrite(wpath.buf, &chdir_cb, NULL);
  MUTEX_UNLOCK(&big_giant_lock);
  
  return retval;
}

/* ------
 * getcwd
 * ------
 */

char *tracker_getcwd(char *buf, size_t size)
{
  MUTEX_UNLOCK(&big_giant_lock);
  if (size == 0) {
    errno = EINVAL;
    return NULL;
  }
  
  if (size < cwd.len + 1) {
    errno = ERANGE;
    return NULL;
  }
  if (!buf) {
    buf = malloc(size);
    if (!buf) {
      errno = ENOMEM;
      return NULL;
    }
  }
  path_terminate(&cwd);
  memcpy(buf, cwd.buf, cwd.len+1);
  MUTEX_UNLOCK(&big_giant_lock);
  return buf; 
}

/* --------
 * getdents
 * --------
 */

static int dirent_cmp_cb(void *l, void *r)
{
  dirent_t *left = (dirent_t *)l;
  dirent_t *right = (dirent_t *)r;
  
  return strcmp(left->d_name, right->d_name);
}

static int getdents_cb(const char *path, const char *vpath, void *cb_data)
{
  tree_node_t **root = (tree_node_t **)cb_data;
  char buf[GETDENTS_BUFSIZE];
  int fd, bufsize;
  
  while (((fd = libc_open(path, O_RDONLY, 0)) < 0) && (errno == EINTR)) {}
  if (fd < 0) {
    return -1;
  }
  do {
    while ((bufsize = libc_getdents(fd, (dirent_t *)buf, GETDENTS_BUFSIZE)) > 0) {
      char *p = buf;
      while (p < (buf + bufsize)) {
        dirent_t *dirent;
        
        dirent = (dirent_t *)p;
        tree_insert(root, &dirent_cmp_cb, dirent, dirent->d_reclen);
        p += dirent->d_reclen;
      }
    }
  } while ((bufsize < 0) && (errno == EINTR));
  libc_close(fd);
  if (bufsize < 0) {
    return -1;
  }
  
  errno = ENOENT;
  return -1;
}

int tracker_getdents(int fd, void *buf, size_t bufsize)
{
  fdinfo_t *fdinfo;
  dirent_t *dirent;
  int nbytes;
  
  // Return an error if the file descriptor is outside the table
  if ((fd < 0) || (fd > fdinfo_tbl_num_sub_pages)) {
    errno = EBADF;
    return -1;
  }
  
  // Find the start of any duplicate chain
  MUTEX_LOCK(&big_giant_lock);
  fd = traverse_dup_chain(fd);
  
  // Check that the record refers to a directory
  fdinfo = get_fd_record(fd);
  if (fdinfo->rec_type != DIRECTORY) {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = ENOTDIR;
    return -1;
  }
  
  // Generate a tree of uniquely named dirents
  if (!fdinfo->rec.dir.dirents_root) {
    tree_node_t *root, *first, *node;
    int off;
    
    root = NULL;
    fdinfo->rec.dir.vpath_buf[fdinfo->rec.dir.vpath_len] = '\0';
    path_rewrite(fdinfo->rec.dir.vpath_buf, &getdents_cb, &root);
    first = tree_linearize(root);
    
    // Recalculate offsets
    node = first;
    off = 0;
    while (node) {
      dirent = (dirent_t *)((char *)node + sizeof(tree_node_t));
      dirent->d_off = off;
      off += dirent->d_reclen;
      node = node->next;
    }
    
    fdinfo->rec.dir.dirents_root = root;
    fdinfo->rec.dir.dirents_first = first;
    fdinfo->rec.dir.dirents_curr = first;
  }
  
  // Write dirents into application buffer
  nbytes = 0;
  while ((fdinfo->rec.dir.dirents_curr) &&
         (dirent = (dirent_t *)((char *)fdinfo->rec.dir.dirents_curr + sizeof(tree_node_t))) &&
         (nbytes + dirent->d_reclen <= bufsize)) {
      memcpy((char *)buf + nbytes, dirent, dirent->d_reclen);
      nbytes += dirent->d_reclen;
      fdinfo->rec.dir.dirents_curr = fdinfo->rec.dir.dirents_curr->next;
  }
  
  if (nbytes || !fdinfo->rec.dir.dirents_curr) {
    MUTEX_UNLOCK(&big_giant_lock);
    return nbytes;
  }
  else {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = EINVAL;
    return -1;
  }
}

#ifdef _ILP32

/* ----------
 * getdents64
 * ----------
 */

static int dirent64_cmp_cb(void *l, void *r)
{
  dirent64_t *left = (dirent64_t *)l;
  dirent64_t *right = (dirent64_t *)r;
  
  return strcmp(left->d_name, right->d_name);
}

static int getdents64_cb(const char *path, const char *vpath, void *cb_data)
{
  tree_node_t **root = (tree_node_t **)cb_data;
  char buf[GETDENTS_BUFSIZE];
  int fd, bufsize;
  
  while (((fd = libc_open64(path, O_RDONLY, 0)) < 0) && (errno == EINTR)) {}
  if (fd < 0) {
    return -1;
  }
  do {
    while ((bufsize = libc_getdents64(fd, (dirent64_t *)buf, GETDENTS_BUFSIZE)) > 0) {
      char *p = buf;
      while (p < (buf + bufsize)) {
        dirent64_t *dirent;
        
        dirent = (dirent64_t *)p;
        tree_insert(root, &dirent64_cmp_cb, dirent, dirent->d_reclen);
        p += dirent->d_reclen;
      }
    }
  } while ((bufsize < 0) && (errno == EINTR));
  libc_close(fd);
  if (bufsize < 0) {
    return -1;
  }
  
  errno = ENOENT;
  return -1;
}

int tracker_getdents64(int fd, void *buf, size_t bufsize)
{
  fdinfo_t *fdinfo;
  dirent64_t *dirent;
  int nbytes;
  
  // Return an error if the file descriptor is outside the table
  if ((fd < 0) || (fd > fdinfo_tbl_num_sub_pages)) {
    errno = EBADF;
    return -1;
  }
  
  // Find the start of any duplicate chain
  MUTEX_LOCK(&big_giant_lock);
  fd = traverse_dup_chain(fd);
  
  // Check that the record refers to a directory
  fdinfo = get_fd_record(fd);
  if (fdinfo->rec_type != DIRECTORY) {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = ENOTDIR;
    return -1;
  }
  
  // Generate a tree of uniquely named dirents
  if (!fdinfo->rec.dir.dirents_root) {
    tree_node_t *root, *first, *node;
    int off;
    
    root = NULL;
    fdinfo->rec.dir.vpath_buf[fdinfo->rec.dir.vpath_len] = '\0';
    path_rewrite(fdinfo->rec.dir.vpath_buf, &getdents64_cb, &root);
    first = tree_linearize(root);
    
    // Recalculate offsets
    node = first;
    off = 0;
    while (node) {
      dirent = (dirent64_t *)((char *)node + sizeof(tree_node_t));
      dirent->d_off = off;
      off += dirent->d_reclen;
      node = node->next;
    }
    
    fdinfo->rec.dir.dirents_root = root;
    fdinfo->rec.dir.dirents_first = first;
    fdinfo->rec.dir.dirents_curr = first;
  }
  
  // Write dirents into application buffer
  nbytes = 0;
  while ((fdinfo->rec.dir.dirents_curr) &&
         (dirent = (dirent64_t *)((char *)fdinfo->rec.dir.dirents_curr + sizeof(tree_node_t))) &&
         (nbytes + dirent->d_reclen <= bufsize)) {
      memcpy((char *)buf + nbytes, dirent, dirent->d_reclen);
      nbytes += dirent->d_reclen;
      fdinfo->rec.dir.dirents_curr = fdinfo->rec.dir.dirents_curr->next;
  }
  
  if (nbytes || !fdinfo->rec.dir.dirents_curr) {
    MUTEX_UNLOCK(&big_giant_lock);
    return nbytes;
  }
  else {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = EINVAL;
    return -1;
  }
}

#endif
