/*
 * utils.c
 * Copyright (C) 2005 Robin KAY. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 */

#include "compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>
#include <limits.h>
#include <unistd.h>

#include "utils.h"
#include "mallory.h"

static int next_path(path_buf_t *wpath, char type, char **pptr)
{
  int len;
  
  if (**pptr != type) {
    return 0;
  }
  
  (*pptr)++;
  len = (int)strtol(*pptr, pptr, 10);
  if (wpath) {
    if (!path_append(wpath, *pptr, len)) {
      return 0;
    }
  }
  (*pptr) += len;
  
  return len;
}

path_buf_t *path_clear(path_buf_t *wpath)
{
  if (!wpath) {
    errno = EFAULT;
    return NULL;
  }
  
  wpath->len = 0;
  return wpath;
}

path_buf_t *path_terminate(path_buf_t *wpath)
{
  if (!wpath) {
    errno = EFAULT;
    return NULL;
  }
  
  wpath->buf[wpath->len] = '\0';
  return wpath;
}

path_buf_t *path_append(path_buf_t *wpath, const char *path, int path_len)
{
  if (!wpath || !path) {
    errno = EFAULT;
    return NULL;
  }
  if (wpath->len + path_len > PATH_MAX - 1) {
    errno = ENAMETOOLONG;
    return NULL;
  }
  memcpy(wpath->buf + wpath->len, path, path_len);
  wpath->len += path_len;
  return wpath;
}

path_buf_t *path_absolute(path_buf_t *wpath, const char *path)
{
  const char *start_p, *cur_p;
  
  if (!wpath || !path) {
    errno = EFAULT;
    return NULL;
  }
  if (path[0] == '/') {
    if (!path_append(path_clear(wpath), "/", 1)) {
      return NULL;
    }
  }
  
  start_p = cur_p = path;
  for (;;) {
    if (*cur_p == '/' || *cur_p == '\0') {
      int len = cur_p - start_p;
      
      if (strncmp(start_p, ".", len) == 0) {
      }
      else if (strncmp(start_p, "..", len) == 0) {
        while ((wpath->len > 1) &&
               (wpath->buf[wpath->len - 1] != '/')) {
          wpath->len--;
        }
        while ((wpath->len > 1) &&
               (wpath->buf[wpath->len - 1] == '/')) {
          wpath->len--;
        }
      }
      else {
        if (wpath->buf[wpath->len - 1] != '/') {
          if (!path_append(wpath, "/", 1)) {
            return NULL;
          }
        }
        if (!path_append(wpath, start_p, len)) {
          return NULL;
        }
      }
      if (*cur_p == '\0') {
        break;
      }
      while (*cur_p == '/') {
        cur_p++;
      }
      start_p = cur_p;
    }
    else {
      cur_p++;
    }
  }
  
  return wpath;
}

int path_rewrite(const char *path, path_rewrite_cb_t cb, void *cb_data)
{
  char *rule, *best_rule;
  const char *best_path_suffix;
  int best_depth, best_path_suffix_len;
  char wpath_buf[PATH_MAX];
  path_buf_t wpath;
  
  // Iterate through the rule set to find the closest matching path
  rule = rules_env;
  best_rule = NULL;
  best_depth = 0;
  wpath.buf = wpath_buf;
  
  while (next_path(path_clear(&wpath), ':', &rule)) {
    const char *basep = wpath.buf;
    const char *pathp = path;
    int match, depth;
    
    path_terminate(&wpath);
    
    match = 1;
    depth = 0;
    for (;;) {
      if (*basep == '/' || *basep == '\0') {
        depth++;
        while (*basep == '/') {
          basep++;
        }
        while (*pathp == '/') {
          pathp++;
        }
        if (*basep == '\0') {
          break;
        }
      }
      else if (*basep == *pathp){
        match = 1;
        basep++;
        pathp++;
      }
      else {
        match = 0;
        break;
      }
    }
    
    if (match) {
      if (depth > best_depth) {
        best_rule = rule;
        best_path_suffix = pathp;
        best_path_suffix_len = strlen(best_path_suffix);
        best_depth = depth;
      }
    }
    
    // Skip all the rule's target paths
    while (next_path(NULL, '+', &rule)) {}
  }
  
  // If there's no matching rule then use the original path
  if (!best_rule) {
    return cb(path, path, cb_data);
  }
  
  // Iterate through the rule's target paths until the callback succeeds.
  rule = best_rule;
  while (next_path(path_clear(&wpath), '+', &rule)) {
    int retval;
    
    if (!(path_append(&wpath, "/", 1) &&
          path_append(&wpath, best_path_suffix, best_path_suffix_len))) {
      return -1;
    }
    path_terminate(&wpath);
    
    retval = cb(wpath.buf, path, cb_data);
    if ((retval >= 0) || (errno != ENOENT)) {
      return retval;
    }
  }
  
  return -1;
}

void tree_insert(
  tree_node_t **nodep, tree_cmp_cb_t cmp_cb, void *data, size_t data_size)
{
  tree_node_t *node;
  
  node = *nodep;
  if (node) {
    int cmp = cmp_cb(data, (char *)node + sizeof(tree_node_t));
    if (cmp > 0) {
      tree_insert(&(node->left), cmp_cb, data, data_size);
    }
    else if (cmp < 0) {
      tree_insert(&(node->right), cmp_cb, data, data_size);
    }
  }
  else {
    node = malloc(sizeof(tree_node_t) + data_size);
    node->left = NULL;
    node->right = NULL;
    node->next = NULL;
    memcpy((char *)node + sizeof(tree_node_t), data,  data_size);
    *nodep = node;
  }
}

void tree_cleanup(tree_node_t **nodep)
{
  tree_node_t *node;
  
  node = *nodep;
  *nodep = NULL;
  if (node) {
    tree_cleanup(&(node->left));
    tree_cleanup(&(node->right));
    free(node);
  }
}

static void linearize(tree_node_t *node, tree_node_t ***prev_fref)
{
  if (node->left) {
    linearize(node->left, prev_fref);
  }
  **prev_fref = node;
  *prev_fref = &(node->next);
  if (node->right) {
    linearize(node->right, prev_fref);
  }
}

tree_node_t *tree_linearize(tree_node_t *node)
{
  tree_node_t *first, **prev_fref;
  
  prev_fref = &first;
  linearize(node, &prev_fref);
  
  return first;
}
